import pytest
import pathlib
import src.adash as _


class TestDownload:
    @pytest.fixture()
    def path(self):
        _p = "tmp/hello.js"
        yield _p
        _p = pathlib.Path(_p)
        if _p.exists():
            pathlib.Path(_p).unlink()

    def test_download(self, path):
        url = "https://cdn.jsdelivr.net/npm/js-hello-world@1.0.0/helloWorld.js"
        p = pathlib.Path(path)
        assert not p.exists()
        assert _.download(url, p) == 1
        assert _.download(url, path) == 0
        text = p.read_text()
        assert 'console.log("Hello World");' in text

    def test_download_unknown(self, path):
        url = "https://www.google.com/example"
        with pytest.warns(UserWarning) as record:
            assert _.download(url, "down_test.html") == 0
            assert record[0].message.args[0] == "404: Not Found"
            assert record[1].message.args[0] == "url: https://www.google.com/example"


class TestJson:
    @pytest.fixture()
    def path(self):
        _p = "tmp/sample.json"
        yield _p
        pathlib.Path(_p).unlink()

    @pytest.fixture()
    def ppath(self):
        _p = pathlib.Path("sample.json")
        yield _p
        pathlib.Path(_p).unlink()

    def test_json(self, path):
        assert _.json_write({"a": 1}, path) == 1
        assert _.json_write({"a": 1}, path, overwrite=True) == 1
        assert _.json_read(path) == {"a": 1}
        assert _.json_write({"a": 1}, path) == 0

    def test_json_pathlib(self, ppath):
        assert _.json_write({"a": 1}, ppath) == 1
        assert _.json_write({"a": 1}, ppath, overwrite=True) == 1
        assert _.json_read(ppath) == {"a": 1}
        assert _.json_write({"a": 1}, ppath) == 0


class TestCat:
    @pytest.fixture()
    def path(self):
        _p = "tmp/hello.js"
        _.json_write({"a": 1}, _p)
        yield _p
        _p = pathlib.Path(_p)
        if _p.exists():
            pathlib.Path(_p).unlink()

    def test_cat(self, path):
        s = _.cat("tmp/hello.js")
        assert s == '{"a": 1}'

    def test_cat_is_none(self, path):
        s = _.cat("tmp/unknown.js")
        assert s is None



def allkeys(d: dict) -> list:
    """ネストした要素も含めたdict.keys

    Args:
        d (dict): dict

    Returns:
        list: keys

    Example:
        >>> d = {'name': {'first': 'John', 'last': 'Smith'}, 'age': 36}
        >>> allkeys(d)
        ['name.first', 'name.last', 'age']

    Note:
        https://qiita.com/ainn_lll/items/e898b7bc8bfc4afdb445
    """
    keys = []
    for parent, children in d.items():
        if isinstance(children, dict):
            keys += [parent + "." + child for child in allkeys(children)]
        else:
            keys.append(parent)
    return keys

def find(lst, predicate):
    """
    Listの中から、predicateの条件に合致する最初の要素を返す。
    
    Args:
        lst (list): 検索対象のリスト。
        predicate (Union[dict, str, Callable]): 検索条件。辞書、文字列、または関数（ラムダ式を含む）のいずれか。
    
    Returns:
        dict or None: 条件に合致する最初の要素。合致する要素がない場合はNone.
        
    Example:
        >>> users = [
        ...     { 'user': 'barney',  'age': 36, 'active': True },
        ...     { 'user': 'fred',    'age': 40, 'active': False },
        ...     { 'user': 'pebbles', 'age': 1,  'active': True }
        ... ]
        >>> find(users, {'age': 1, 'active': True})
        {'user': 'pebbles', 'age': 1, 'active': True}
        >>> find(users, 'active')
        {'user': 'barney', 'age': 36, 'active': True}
        >>> find(users, lambda x: x['age'] < 40 or not x['active'])
        {'user': 'barney', 'age': 36, 'active': True}
    """
    if isinstance(predicate, dict):
        for item in lst:
            if all(item[key] == val for key, val in predicate.items()):
                return item
    elif isinstance(predicate, str):
        for item in lst:
            if isinstance(item, dict):
                if item.get(predicate) is True:
                    return item
            else:
                if item == predicate:
                    return item
    else:
        for item in lst:
            if predicate(item):
                return item
    return None



from typing import Any, Union
import urllib.request
import urllib.error
import pathlib
import warnings
import os
import time
import json


def download(url, save_path, sleep=1):
    """
    urlからダウンロードする。save_pathが存在するなら何もしない。
    Args:
        url (str): url
        save_path (str): save path
        sleep (int): 待機時間(秒)
    Returns:
        int: 成功なら1,urlがnotfoundまたはファイルが既に存在するなら0
    """
    save_path = pathlib.Path(save_path)
    if save_path.exists():
        return 0
    time.sleep(sleep)
    os.makedirs(save_path.parent, exist_ok=True)
    req = urllib.request.Request(url)
    try:
        with urllib.request.urlopen(req) as res:
            save_path.write_bytes(res.read())
        return 1
    except urllib.error.HTTPError as e:
        warnings.warn(f"{e.code}: {e.reason}")
        warnings.warn(f"url: {url}")
        return 0


def json_write(obj: Union[dict, list], file_path: str, overwrite: bool = False) -> int:
    """dictをjsonファイルに書き込む

    Args:
        obj (Union[dict, list]): json化するオブジェクト
        file_path (str): ファイルパス
        overwrite (bool, optional): Trueで上書きを許す. Defaults to False.

    Returns:
        int: 成功なら1が返る

    Example:
        _.json_write({"a": 1}, "path/to/sample.json")
        _.json_write({"a": 1}, "path/to/sample.json", overwrite=True)
    """
    file_path = pathlib.Path(file_path)
    if not overwrite and file_path.exists():
        return 0
    os.makedirs(file_path.parent, exist_ok=True)
    file_path.write_text(json.dumps(obj))
    return 1


def json_read(file_path: str) -> Any:
    """jsonファイルを読み込んでdict化する

    Args:
        file_path (str): ファイルパス

    Returns:
        Any: dictもしくはlist

    Example:
        _.json_read("path/to/sample.json")
    """
    file_path = pathlib.Path(file_path)
    return json.loads(file_path.read_text())


def cat(file_path: str) -> str:
    """pathlib.Path().read_textのshortcut

    Args:
        file_path (str): filepath

    Returns:
        str: file内の文字列

    Example:
        >>> cat('unknown.txt')

    """
    file_path = pathlib.Path(file_path)
    if file_path.is_file():
        return file_path.read_text()
    return None



from datetime import timedelta, datetime


def date_list(day: int = 30, start: str = "today", fmt: str = "%Y-%m-%d", rev=False):
    """n日前までの日付文字列をlist化
    Args:
        day (int): 日数
        start: (str): 開始の日、例えば`2021-12-01`とすればこの日からn日前までのlistを返す
        fmt (str): strftimeで使われるformat
        rev (bool): 結果をriverseする(defalt=Falseでは正順)
    Returns:
        list: n日分の日付文字列

    Example:
        >>> date_list(3, '2020-12-1')
        ['2020-11-29', '2020-11-30', '2020-12-01']
        >>> date_list(3, '20201201', '%Y%m%d')
        ['20201129', '20201130', '20201201']
        >>> date_list(3, '2020-12-01', rev=True)
        ['2020-12-01', '2020-11-30', '2020-11-29']

    Note:
        https://docs.python.org/ja/3/library/datetime.html#strftime-strptime-behavior
    """
    d = datetime.now()
    if start != "today":
        d = datetime.strptime(start, fmt)
    li = [d.strftime(fmt)]
    for _ in range(day - 1):
        d -= timedelta(days=1)
        li.append(d.strftime(fmt))
    if not rev:
        li.reverse()
    return li



import src.adash as _


class TestAllkeys:
    def test_01(self):
        d = {"name": {"first": "John", "last": "Smith"}, "age": 36}
        res = _.allkeys(d)
        assert res == ["name.first", "name.last", "age"]

    def test_02(self):
        # https://json.org/example.html
        d = {
            "glossary": {
                "title": "example glossary",
                "GlossDiv": {
                    "title": "S",
                    "GlossList": {
                        "GlossEntry": {
                            "ID": "SGML",
                            "SortAs": "SGML",
                            "GlossTerm": "Standard Generalized Markup Language",
                            "Acronym": "SGML",
                            "Abbrev": "ISO 8879:1986",
                            "GlossDef": {
                                "para": "A meta-markup language, used to create markup languages such as DocBook.",
                                "GlossSeeAlso": ["GML", "XML"],
                            },
                            "GlossSee": "markup",
                        }
                    },
                },
            }
        }
        assert _.allkeys(d) == [
            "glossary.title",
            "glossary.GlossDiv.title",
            "glossary.GlossDiv.GlossList.GlossEntry.ID",
            "glossary.GlossDiv.GlossList.GlossEntry.SortAs",
            "glossary.GlossDiv.GlossList.GlossEntry.GlossTerm",
            "glossary.GlossDiv.GlossList.GlossEntry.Acronym",
            "glossary.GlossDiv.GlossList.GlossEntry.Abbrev",
            "glossary.GlossDiv.GlossList.GlossEntry.GlossDef.para",
            "glossary.GlossDiv.GlossList.GlossEntry.GlossDef.GlossSeeAlso",
            "glossary.GlossDiv.GlossList.GlossEntry.GlossSee",
        ]

    def test_03(self):
        assert _.allkeys({"a": 1, "b": 1}) == ["a", "b"]


class TestFind:
    def test_01(self):
        users = [
            {"user": "barney", "age": 36, "active": True},
            {"user": "fred", "age": 40, "active": False},
            {"user": "pebbles", "age": 1, "active": True},
        ]
        assert _.find(users, {"age": 1, "active": True}) == {
            "user": "pebbles",
            "age": 1,
            "active": True,
        }

    def test_02(self):
        users = [
            {"user": "barney", "age": 36, "active": True},
            {"user": "fred", "age": 40, "active": False},
            {"user": "pebbles", "age": 1, "active": True},
        ]
        assert _.find(users, "active") == {"user": "barney", "age": 36, "active": True}

    def test_03(self):
        users = [
            {"user": "barney", "age": 36, "active": True},
            {"user": "fred", "age": 40, "active": False},
            {"user": "pebbles", "age": 1, "active": True},
        ]
        assert _.find(users, lambda x: x["age"] < 40 or not x["active"]) == {
            "user": "barney",
            "age": 36,
            "active": True,
        }

    def test_04(self):
        items = ["apple", "banana", "cherry", "date"]
        assert _.find(items, lambda x: x.startswith("b")) == "banana"

    def test_05(self):
        items = ["apple", "banana", "cherry", "date"]
        assert _.find(items, "grape") is None

    def test_06(self):
        users = [
            {"user": "barney", "age": 36, "active": True},
            {"user": "fred", "age": 40, "active": False},
            {"user": "pebbles", "age": 1, "active": True},
        ]
        assert _.find(users, {"age": 100, "active": False}) is None



from decimal import Decimal, ROUND_HALF_UP, InvalidOperation, DivisionByZero


def _proportion(base: float, target: float, round_number: str) -> Decimal:
    """baseとtargetの比率
    計算式: (target - base) / abs(base)
    """
    numerator = Decimal(str(target)) - Decimal(str(base))
    denominator = Decimal(str(base)).copy_abs()
    # return float((numerator / denominator * 100).quantize(Decimal(round_number), ROUND_HALF_UP))
    return numerator / denominator


def _to_float(d: Decimal, round_number: str) -> float:
    return float(d.quantize(Decimal(round_number), ROUND_HALF_UP))


def proportion(base: float, target: float, round_number: str = "0.01") -> float:
    """baseに対してtargetの比率
    前日終値100円に対し今日の終値99円だと-1%
    Example:
        >>> proportion(100, 99)
        -1.0
        >>> proportion(203, 7)
        -96.55
    """
    try:
        p = _proportion(base, target, round_number)
        return _to_float(p * 100, round_number)
    except (InvalidOperation, ZeroDivisionError):
        return float("nan")


def progress_rate(goal: float, progress: float, round_number: str = "0.01") -> float:
    """goalに対して現在(progress)の達成率
    Example:
        >>> progress_rate(50, 40)
        80.0
        >>> progress_rate(50, 12.031454, '0.00001')
        24.06291
    """
    if goal == 0:
        return float("nan")
    try:
        p = _proportion(goal, progress, round_number)
        return _to_float((Decimal("1") + p) * 100, round_number)
    except (InvalidOperation, ZeroDivisionError):
        return float("nan")



from math import isnan

import src.adash as _


class TestReplaceAll:
    def test_01(self):
        s = "3円00銭"
        obj = {"円": ".", "銭": ""}
        assert _.replace_all(s, obj) == "3.00"

    def test_02(self):
        s = "abcabc"
        obj = {"a": "!", "b": "", "c": "?"}
        assert _.replace_all(s, obj) == "!?!?"

    def test_03(self):
        obj = {"[△▲]": "-", "[,、]": ""}
        assert _.replace_all("▲12,345", obj) == _.replace_all("△12、345", obj)


class TestHalfString:
    def test_01(self):
        assert _.to_half_string("１２３４５ａｂｃ") == "12345abc"


class TestToNumber:
    def test_01(self):
        assert _.to_number("123,456") == 123456
        assert _.to_number("-123,456") == -123456
        assert _.to_number("▲123,456") == -123456
        assert _.to_number("△123,456") == -123456
        assert isnan(_.to_number("▼123,456"))
        assert isnan(_.to_number("abc"))
        assert _.to_number("abc", True) is True
        assert _.to_number("abc", False) is False
        assert _.to_number("18円32銭") == 18.32
        assert _.to_number("18円00銭") == 18
        assert _.to_number("18円") == 18
        assert _.to_number("18%") == 18
        assert _.to_number("18％") == 18
        assert _.to_number("１２３，４５６，７８９") == 123456789


class TestToDate:
    def test_01(self):
        td = _.to_date
        assert td("２０１８年５月２７日") == "2018-05-27"
        assert td("2018年1月1日") == "2018-01-01"
        assert td("平成30年 5月11日") == "2018-05-11"
        assert td("令和元年 5月11日") == "2019-05-11"
        assert td("令和 元年 5月12日") == "2019-05-12"
        assert td("令和2年 5月11日") == "2020-05-11"
        assert td("平成３０年５月２７日") == "2018-05-27"
        assert td("昭和３０年５月２７日") == "1955-05-27"
        assert td("大正３年５月２７日") == "1914-05-27"
        assert td("大３年５月２７日") == "1914-05-27"
        assert td("明３年５月２７日") == "1870-05-27"
        assert td("2018-5-27") == "2018-05-27"
        assert td("2018/5/27") == "2018-05-27"
        assert td("20180527") == "2018-05-27"
        assert td("Unknown") is None


class TestSplitUpperCase:
    def test_01(self):
        assert _.split_uppercase("IsJSON") == ["Is", "JSON"]
        assert _.split_uppercase("ILoveYou") == ["ILove", "You"]
        assert _.split_uppercase("NextAccumulatedQ2Duration") == [
            "Next",
            "Accumulated",
            "Q2",
            "Duration",
        ]


class TestHash:
    def test_md5(self):
        assert _.to_md5("test") == "098f6bcd4621d373cade4e832627b4f6"
        assert _.to_md5("test1") == "5a105e8b9d40e1329780d62ea2265d8a"

    def test_sha256(self):
        assert (
            _.to_sha256("test")
            == "9f86d081884c7d659a2feaa0c55ad015a3bf4f1b2b0b822cd15d6c15b0f00a08"
        )
        assert (
            _.to_sha256("test1")
            == "1b4f0e9851971998e732078544c96b36c3d01cedf7caa332359d6f1d83567014"
        )


class TestTextNormalize:
    def test_01(self):
        assert _.text_normalize("ﾊﾝｶｸｶﾀｶﾅｚｅｎｋａｋｕ１２３") == "ハンカクカタカナzenkaku123"
        assert _.text_normalize("①②③１２３") == "①②③123"
        assert _.text_normalize("①②③１２３", exclude_chars="①１") == "①23１23"



import hashlib
import re
import unicodedata
from typing import Any


def replace_all(s: str, obj: dict) -> str:
    """複数のreplace
    Example:
        >>> _obj = {"円": ".", "銭": ""}
        >>> replace_all("3円00銭", _obj)
        '3.00'
        >>> _obj = {"[△▲]": "-", "[,、]": ""}
        >>> replace_all('▲12,345', _obj)
        '-12345'
        >>> replace_all('△12、345', _obj)
        '-12345'
    """
    for key in obj:
        val = obj[key]
        s = re.sub(key, val, s)
    return s


def to_half_string(s: str) -> str:
    """normalize
    Example:
        >>> to_half_string('１２３ＸＹＺ')
        '123XYZ'
    """
    return unicodedata.normalize("NFKC", s)


def to_number(s: str, default: Any = float("NAN")) -> Any:
    """文字列を数値化
    Example:
        >>> to_number('△12,345')
        -12345
        >>> to_number('１２，３４５')
        12345
        >>> to_number('12円34銭')
        12.34
        >>> to_number('98%')
        98
        >>> to_number('abc')
        nan
        >>> to_number('abc', '123')
        '123'
    """
    rep_dict = {"[△▲Δ]": "-", "[,、銭%％]": "", "[円]": "."}
    s = to_half_string(s)
    s = replace_all(s, rep_dict)
    try:
        float(s)
    except ValueError:
        return default
    else:
        if float(s).is_integer():
            return int(float(s))
        return float(s)


def to_date(s: str) -> Any:
    """日付表記を統一する
    Example:
        >>> to_date('２０１８年５月２７日')
        '2018-05-27'
        >>> to_date('令和 元年 5月12日')
        '2019-05-12'
        >>> to_date('大３年５月２７日')
        '1914-05-27'
        >>> to_date('20211010')
        '2021-10-10'
    """
    meiji = "meiji"
    taisyou = "taisyou"
    syouwa = "syouwa"
    heisei = "heisei"
    reiwa = "reiwa"
    s = to_half_string(s)
    s = replace_all(
        s,
        {
            "[年月/]": "-",
            r"[日\s]": "",
            "[元]": "1",
            r"(明治|明)": meiji,
            r"(大正|大)": taisyou,
            r"(昭和|昭)": syouwa,
            r"(平成|平)": heisei,
            r"(令和|令)": reiwa,
        },
    )
    # 数字8桁ならハイフン挿入
    m = re.match(r"^\d{8}$", s)
    if m:
        s = re.sub(r"(\d{4})(\d{2})(\d{2})", r"\1-\2-\3", s)
    m = re.match(r"(\D*)(\d*)-(\d*)-(\d*)", s)
    if m:
        era, year, month, day = m.groups()
        if era:
            if era == meiji:
                year = int(year) + 1868 - 1
            elif era == taisyou:
                year = int(year) + 1912 - 1
            elif era == syouwa:
                year = int(year) + 1926 - 1
            elif era == heisei:
                year = int(year) + 1989 - 1
            elif era == reiwa:
                year = int(year) + 2019 - 1
            year = str(year)
        month = f"0{month}" if len(month) == 1 else month
        day = f"0{day}" if len(day) == 1 else day
        return f"{year}-{month}-{day}"
    return None


def split_uppercase(s: str) -> list:
    """UpperCaseを分割
    Example:
        >>> split_uppercase('NextAccumulatedQ2Duration')
        ['Next', 'Accumulated', 'Q2', 'Duration']
    """
    return re.findall(r"[A-Z]+[a-z0-9]*", s)


def to_md5(s: str):
    """文字列をmd5化
    Example:
        >>> to_md5('test')
        '098f6bcd4621d373cade4e832627b4f6'
    """
    return hashlib.md5(s.encode()).hexdigest()


def to_sha256(s: str):
    """文字列をsha256化
    Example:
        >>> to_sha256('test')
        '9f86d081884c7d659a2feaa0c55ad015a3bf4f1b2b0b822cd15d6c15b0f00a08'
    """
    return hashlib.sha256(s.encode()).hexdigest()


def text_normalize(text: str, exclude_chars: str = "①②③④⑤⑥⑦⑧⑨⑩⑪⑫⑬⑭⑮⑯⑰⑱⑲⑳") -> str:
    """文字列の正規化
    Example:
        >>> text_normalize("ﾊﾝｶｸｶﾀｶﾅｚｅｎｋａｋｕ１２３")
        'ハンカクカタカナzenkaku123'
        >>> text_normalize("①②③１２３")
        '①②③123'
        >>> text_normalize("①②③１２３", exclude_chars="①１")
        '①23１23'
    """
    normalized_text = []
    for char in text:
        if char in exclude_chars:
            normalized_text.append(char)
        else:
            normalized_text.append(unicodedata.normalize("NFKC", char))
    return "".join(normalized_text)



"""
.. include:: ../../README.md
"""
from .collection_util import allkeys, find
from .date_util import date_list
from .file_util import cat, download, json_read, json_write
from .proportion import progress_rate, proportion
from .string_util import (
    replace_all,
    split_uppercase,
    text_normalize,
    to_date,
    to_half_string,
    to_md5,
    to_number,
    to_sha256,
)
